#ifndef BITMAP_H
#define BITMAP_H

#include <alleggl.h>
#include <allegro.h>
#include <string>

// Represents a tint.  Colors range from 0 to 1 where 1 is draw this color
// and 0 is do not draw this color.
//
// For example to apply a blue tint you would have:
// r = 1, g = 0, b = 0
//
// The alpha value specifies at what opacity to draw the texture.  1 implies
// draw the texture as it is while .5 implies draw half transparent and 0
// implies do not draw at all.
struct Tint {
    
    float r, g, b, a;
    
    // True if the blit should be done with additive blending.
    bool additive;
    
    Tint(float red = 1, float green = 1, float blue = 1, float alpha = 1, bool add = false)
        : r(red), g(green), b(blue), a(alpha), additive(add)
    {
        
    }
};

// Contains a bitmap or raster.
class Bitmap {
public:
    
    // Draws a gradient to the screen
    static void gradient(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft);
    
    // Draws a line to the screen
    static void line(int x1, int y1, int x2, int y2, int width, const Tint &one, const Tint &two);
    
    // Sets a clipping rectangle starting with x, y specifying the upper
    // left hand corner of the rectangle.  scissorStart(0, 0, 1, 1) would allow
    // drawing on only the upper left pixel on the screen.  Be sure to call
    // scissorEnd to disable clipping again when you're done with it.
    static void scissorStart(int x, int y, int w, int h);
    
    // Disables scissor clipping again
    static void scissorEnd();
    
    // Sets the screen shake values.  All blitting routines will be drawn at the
    // bitmap's x / y plus xoffset / offset respectively.
    static void setShake(int xoffset, int yoffset);
    
    // Gets the current screen shake values
    static void getShake(int *xoffset, int *offset);
    
    // Saves the screen to the given filename.  The type of the file is
    // determined by its extension.  Supports all extension formats
    // supported by allegro
    static void screenshot(const std::string &filename);
    
    Bitmap();
    
    // Copies src to a texture or stores the pointer depending on driver settings
    Bitmap(BITMAP *src);
    
    Bitmap(const Bitmap &src);
    
    Bitmap &operator =(const Bitmap &source);
    
    ~Bitmap();
    
    // Must be called before any bitmaps can be instanciated
    static void init();
    
    // Must be called after all bitmaps have been freed
    static void shutdown();
    
    // Unloads all Bitmaps.  You must call reloadAll afterwards
    // to prevent the program from crashing.
    static void unloadAll();
    
    // Reloads all Bitmaps according to their source bitmaps.
    // This will only work if the BITMAP*s used to make all
    // the Bitmaps are still valid, if any have been freed
    // the program will segfault.
    //
    // You must call unloadAll before using this function
    // to prevent crashes
    static void reloadAll();
    
    // Locks the bitmap mutex blocking all bitmap loads
    static void lockLoading(); // This function does nothing but is here for compatibility reasons
    
    // Unlocks the bitmap mutex allowing bitmaps loads again
    static void unlockLoading(); // This function does nothing but is here for compatibility reasons
    
    // Draws to the screen with settings supplied to tint.  See the
    // comments on the Tint class for more information on this is used.
    void blitToScreen(int x, int y, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Draws to the screen with settings supplied to tint.  See the
    // comments on the Tint class for more information on this is used.
    void blitToScreend(double x, double y, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Draws a circle to the screen with x / y being the center of the circle,
    // radius being the distance from the center to the edge of the circle, and
    // angle1 being where to start drawing the circle and angle2 where to stop
    // drawing the circle, the restult is a pac-man like cutaway circle.
    //
    // Look at this example diagram:
    // 1 stands for angle1, and 2 stands for angle2
    /*
               ....................1
         ...........................
       .............................
      ..............................
     ...............................
     ...............................
     ...............................
      .................................
       ....................................
         .....................................
            ......................................
               ......................................2
    */
    // The dots represent the partial circle that will be drawn.
    void circleBlitToScreen(int x, int y, int radius, float angle1, float angle2, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Draws to the screen with settings supplied to tint.  See the
    // comments on the Tint class for more information on this is used.
    // *offset are percent of bitmap that the tiling starts.
    // 
    // For example to start tiling in the center of the bitmap, pass .5 and
    // .5 for xoffset and yoffset.
    void tileBlitToScreen(int x, int y, float w, float h, float xoffset = 1, float yoffset = 1, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Draws to the screen rotated clock-wise by rotation and scaled to
    // width * wchange by height * hchange.  A complete rotation is a value of
    // 1.  See the comments on the Tint class for more information on how to
    // use the tint parameter.
    void specialBlitToScreen(int x, int y, float wchange, float hchange, float rotation, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Version of specialBlitToScreen that takes doubles for the x and y coords.
    void specialBlitToScreend(double x, double y, float wchange, float hchange, float rotation, const Tint &tint = Tint(), bool disableBlending = false);
    
    // Draws to the screen rotating bitmap around point bitmapX, bitmapY.  This
    // point on the bitmap is matched to screenX, screenY on the screen.  Then
    // the bitmap is rotated by the rotation variable.  A rotation value of 1
    // is a complete rotation.
    //
    // Todo: Implement this function.
    // void pivotBlitToScreen(int screenX, int screenY, int bitmapX, int bitmapY, float rotation, const Tint &tint = Tint());
    
    int getw() { return w; }
    int geth() { return h; }
    
    // Returns the texture if the rendering mode is hardware or an invalid texture number
    // if its software.
    GLuint getTexture();
    
    // Returns the bitmap if the rendering mode is software or NULL if it is hardware
    BITMAP *getBitmap();
    
private:
    
    // True if this instance used as parameter to a cctor
    mutable bool copiedFrom;
    
    int w, h;
    BITMAP *bitmap;
    bool isInVram;
    GLuint texture;
    
    // The BITMAP* source used to make this Bitmap
    BITMAP *src;
    
    static void reloadAllHardware();
    static void gradientHardware(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft);
    static void lineHardware(int x1, int y1, int x2, int y2, int width, const Tint &left, const Tint &right);
    static void scissorStartHardware(int x, int y, int w, int h);
    static void scissorEndHardware();
    static void screenshotHardware(const std::string &filename);
    BITMAP *getSrcHardware();
    void loadBitmapHardware(BITMAP *src);
    void unloadBitmapHardware();
    void blitToScreenHardware(int x, int y, const Tint &tint, bool disableBlending);
    void blitToScreenHardwared(double x, double y, const Tint &tint, bool disableBlending);
    void circleBlitToScreenHardware(int x, int y, int radius, float angle1, float angle2, const Tint &tint, bool disableBlending);
    void tileBlitToScreenHardware(int x, int y, float w, float h, float xoffset, float yoffset, const Tint &tint, bool disableBlending);
    void specialBlitToScreenHardware(int x, int y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending);
    void specialBlitToScreenHardwared(double x, double y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending);
    
    static void gradientSoftware(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft);
    BITMAP *getSrcSoftware();
    void loadBitmapSoftware(BITMAP *src);
    void unloadBitmapSoftware();
    void blitToScreenSoftware(int x, int y, const Tint &tint, bool disableBlending);
    void blitToScreenSoftwared(double x, double y, const Tint &tint, bool disableBlending);
    void tileBlitToScreenSoftware(int x, int y, float w, float h, float xoffset, float yoffset, const Tint &tint, bool disableBlending);
    void specialBlitToScreenSoftware(int x, int y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending);
};

#endif
