#ifndef GUISCREEN_H
#define GUISCREEN_H

#include "datafile.h"
#include "widget.h"
#include "animation.h"
#include <map>
#include <string>
#include <queue>
#include <list>
#include <sstream>

class GuiScreen {
public:
    
    // Pointer to the active GuiScreen.  Set to 'this' at the start of poll
    // and load.
    static GuiScreen *active;
    
    // Searches the specified datafile for the filename first, if failing
    // there will search the current working directory.  Passing NULL as
    // the datafile will skip the search for filename in the datafile.
    GuiScreen(const std::string &filename, datafile_t* = 0);
    
    ~GuiScreen();
    
    // Called when its time to draw the screen.
    void draw(const Tint &tint = Tint());
    
    // Returns OK for normal status and QUIT when this screen is done and has shutdown.
    // Must be called constantly but not nessicarly at a constant rate.
    void poll();
    
    // Returns true if there is an action waiting that will be returned by getAction.
    bool actionWaiting();
    
    // Returns the name of an action to be performed.  This is generated by the script that is read
    // in when certain events occur (ie "gamestart").  Do not call this method unless actionWaiting
    // returns true.  commander is a pointer to a pointer to a widget that will be set to a pointer
    // to the widget that caused this action to happen.  In english, it lets you find out what Widget
    // caused this action.
    std::string getAction(Widget **commander = 0);
    
    // Generates an action to be returned by getAction.  Also causes actionWaiting to return
    // true.  sender is a pointer to the Widget that caused this action to happen and is what
    // commander is set to by getAction.
    void generateAction(const std::string &action, Widget *sender);
    
    // Returns true if this GuiScreen should have a window drawn behind it.
    bool hasWindow();
    
    // Finds a widget with a given key set to a given value or NULL
    // if one cannot be found.
    Widget *findByKey(const std::string &key, const std::string &value);
    
    // Sets the cursor to the given image in the datafile.  An empty string sets the default
    // cursor
    void setCursor(const std::string &path);
    
    // Specify where the upper left hand corner of the screen is.  All widgets are passed
    // these values when they're drawn and they're added to the widgets' position.  They're
    // also used to calculate the position of the widgets.
    int xstart, ystart;
    
    // The width and height of this guiscreen.
    int w, h;
    
    // The title of this GuiScreen
    std::string title;
    
    // All the widgets in this GuiScreen.
    std::list<Widget*> widgets;
    
private:
    
    std::queue< std::pair<std::string, Widget*> > actions;
    
    Widget *hasHover, *hasFocus;
    bool oldClick;
    int old_mouse_x, old_mouse_y;
    
    struct EventInfo {
        EventRecieverAsync events;
        std::string keyAction;
        std::string focusAction;
        std::string unfocusAction;
        std::string hoverAction;
        std::string leaveAction;
        std::string clickAction;
        std::string releaseAction;
        
        // The widget that sends the events that the 'events' attribute recieves.
        Widget *widget;
    };
    
    std::list<EventInfo> eventRecievers;
    
    animation defPointer;
    animation *curPointer;
    
    std::map<std::string, animation*> cursors;
    
    bool windowWidgetFound;
    
    // Loads the GuiScreen through the specified stream.
    void load(std::istream &instream);
};

#endif
