#ifndef INPUT_H
#define INPUT_H

#include <string>
#include <queue>
#include <allegro.h>

/* Input
 *  Interface to the user input.  An instace of Input alert's the class's
 *  user of the keyboard events it specifies that it would like to
 *  recieve.  If another input instance requests to recieve input
 *  of an event already being recieved the latter request will be granted
 *  until either the event is un-requested or the instance gets deleted or
 *  goes out of scope.  Once this happens the former request will be granted
 *  again.  This system can handle any number of Input instances.
 *
 *  Note this class uses keypressed and readkey internaly so make sure you
 *  do NOT use these functions anywhere in the program if there are also
 *  Input instances in use.
 */
class Input {
public:
    
    // Requests a character to be recieved.  The actual character can be
    // recieved by getCharacter.  See getCharacter's comments before using
    // the method.
    void requestCharacter(char character);
    
    // Requests multiple characters at once to be recieved.  The actual
    // characters will be recieved by getCharacter one at a time.  See
    // getCharacter's comments before using the method.
    void requestCharacters(const std::string &characters);
    
    // Stops requesting a character that was previously requested with
    // requestCharacter or requestCharacters.  If a character is requested
    // multiple times it must be unrequested the same number of times.
    void unrequestCharacter(char character);
    
    // Stops requesting multiple characters that were previously requested with
    // requestCharacter or requestCharacters.  If a character is requested
    // multiple times it must be unrequested the same number of times.
    void unrequestCharacters(const std::string &characters);
    
    // Returns true if it is safe to call getCharacter.
    bool characterWaiting();
    
    // Gets the next character waiting on the input buffer.  Note that by
    // default this function will return nothing.  To make it return characters
    // as they come in you must request the characters through requestCharacter
    // and requestCharacters.
    //
    // It is unsafe to call this function if characterWaiting returns false.
    char getCharacter();
    
    // Requests an allegro key-code to be recieved.  Some examples of allegro
    // key-codes are:
    // KEY_ESC, KEY_BACKSPACE, KEY_ENTER, KEY_A, KEY_1
    // See allegro's documentation for a complete list of key-codes.
    // The actual key-codes will be recieved by getKeycode.  See
    // getKeycode's comments before using the method.
    void requestKeycode(int keycode);
    
    // Stops this instace requesting an allegro key-code that was previously
    // requested with requestKeycode.  If an allegro key-code is requested
    // multiple times it must be unrequested the same number of times.
    void unrequestKeycode(int keycode);
    
    // Returns true if it is safe to call getKeycode.
    bool keycodeWaiting();
    
    // Gets the next Allegro key-code waiting on the input buffer.  Note that by
    // default this function will return nothing.  To make it return characters
    // as they come in you must request the characters through requestKeycode.
    //
    // It is unsafe to call this function if keycodeWaiting returns false.
    int getKeycode();
    
    // The dtor must un-request all requested input events.
    ~Input();

private:
    
    std::queue<char> characters;
    std::queue<int>  keycodes;
    
    // This function takes the input from the keyboard and passes it to the
    // Input instances who have most recently requested the input values.
    //
    // This function is called internaly by keycodeWaiting and characterWaiting.
    static void getInput();
};

#endif
