#ifndef SOCKET_H
#define SOCKET_H

#include <string>
#include <exception>

#ifdef WIN32
#ifdef ALLEGRO_H
#include <winalleg.h>
#else
#include <winsock.h>
#endif
#else
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#undef INVALID_SOCKET
#define INVALID_SOCKET -1
#undef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

namespace net {
    
    /*** IMPORTANT ***/
    // README: Before you can use any of the functions provided here, you must
    // call init.  When you are done using these routines you must call shutdown.
    
    void init();
    void shutdown();
    
    // Returns the errno value.  Used to determine the type error that
    // occured when a recent function failed.
    int getErrno();
    
    /* Various networking utility functions */
    // Returns a string describing what happened to cause the recent error
    std::string getError(int err = getErrno());
    
    // Sets 'error' to describe what happened to cause the recent error
    void getError(std::string *error, int err = getErrno());
    
    // Returns the errno value.  Used to determine the type error that
    // occured when gethostbyname failed.
    int getHErrno();

    // Returns a string describing what happened when gethostbyname failed.
    std::string getHError(int herrno = getHErrno());
    
    // Sets 'error' to describe what happened when gethostbyname failed.
    void getHError(std::string *error, int herrno = getHErrno());
    
    // Sets hostname to the hostname of the computer.  Returns false on
    // error.  Use getErrno or getError to retrieve the error.
    bool gethostname(std::string *hostname);
    
    // Sets hostname to the hostname of the computer.  Returns false on
    // error.  Use getErrno or getError to retrieve the error.  If status
    // is non-zero it is set to false if an error occured or true if one
    // did not occur.
    std::string gethostname(bool *status = 0);
    
    // Looks up the ip address associated with the hostname 'name' and
    // places it in 'ip'.  Returns false on error.
    //
    // Note: This function's error requires the 'H' variants of the error
    // retrieving function. Use getHErrno or getHError to retrieve the error.
    bool gethostbyname(const std::string &name, std::string *ip);
    
    // Looks up the ip address associated with the hostname 'name' and
    // returns it.  If status is non-zero it is set to false on error or
    // true if there is no error.
    //
    // Note: This function's error requires the 'H' variants of the error
    // retrieving function. Use getHErrno or getHError to retrieve the error.
    std::string gethostbyname(const std::string &name, bool *status = 0);
    
    // Looks up the hostname for the given ip 'ip' and places it in
    // 'hostname'.  Returns false on error.  Use getErrno or getError to
    // retrieve the error.
    bool gethostbyaddr(const std::string &ip, std::string *hostname);
    
    // Looks up the hostname for the given ip 'ip' and returns it.  If
    // status is non-zero sets it to false if an error occured or true if
    // no error occured.  Use getErrno or getError to retrieve the error.
    std::string gethostbyaddr(const std::string &ip, bool *status = 0);
    
    /* A vanilla wrapper around BSD style sockets. */
    // Most methods return false it they have failed and you should call
    // either getErrno or getError to get a description of the error that
    // occured.
    //
    // Do *NOT* create any instances of Socket that are instanciated before
    // main is called.
    //
    // As safe for stl list types as auto_ptr is (invalidation on copied-from).
    class Socket {
    public:
        
        struct Error : public std::exception {};
#ifdef WIN32
        typedef SOCKET Sock;
#else
        typedef int Sock;
#endif
        
        enum Protocol { TCP, UDP };
        
        // Empty socket wrapper (doesnt register a new socket with the kernel).
        Socket();
        
        // Registers a new socket with the kernel using the protocol specified
        // by protocol.
        Socket(Protocol protocol) throw(Error);
        
        // Copies socket, invalidating it if invalidate is true.
        Socket(const Socket &socket, bool invalidate = true);
        
        // Copies socket, invalidating it if invalidate is true. This
        // constructor is for reverse compatibility with BSD sockets.
        Socket(Sock socket, bool invalidate = true);
        
        ~Socket();
        
        // Makes the socket non-blocking
        void makeNonBlocking();
        
        // Binds the socket to the specified port.  If the binding fails bind
        // returns false.  Use getErrno or getError to retrieve the error.
        // Use listen to start the socket listening after is has been bound.
        bool bind(unsigned short port);
        
        // Starts the socket listening on the port specified by bind.  If
        // listening fails listen returns false.  Use getErrno or getError to
        // retrieve the error.  Be sure to call bind before listening.
        bool listen();
        
        // Attempts to connect to the specified ip address specified by ip on
        // the port specified by port.  If the connection attempt fails
        // connectIp returns false. Use getErrno or getError to retrieve the
        // error.
        //
        // connectIp will not work on hostnames (ie google.com).  Use connect
        // if you want to specify hostnames.
        bool connectIp(const std::string &ip, unsigned short port);
        
        // Attempts to connect to the specified ip address specified by ip on
        // the port specified by port.  If the connection attempt fails
        // connectIp returns false. Use getErrno or getError to retrieve the
        // error.
        //
        // connect will work on hostnames (ie google.com).  If you don't want
        // the overhead of calling dns servers use connectIp
        bool connect(const std::string &hostname, unsigned short port);
        
        // Accepts a connection and returns a BSD style socket.  If the
        // accept fails acceptBsd returns INVALID_SOCKET. Use getErrno or
        // getError to retrieve the error.
        //
        // Warning: INVALID_SOCKET's value chagnes depending the OS, so to be
        // portable use this value and *not* -1.
        //
        // Pass a string for ip and / or port to recieve that information about
        // the connection recieved.
        Sock acceptBsd(std::string *ip = 0, unsigned short *port = 0);
        
        // Accepts a connection and returns a Socket for use.  If the
        // accept fails accept returns 0. Use getErrno or getError to retrieve
        // the error.
        //
        // Sets status to false if there was an error or true if no error occured.
        //
        // Pass a string for ip and / or port to recieve that information about
        // the connection recieved.
        Socket accept(bool *status, std::string *ip = 0, unsigned short *port = 0);
        
        // Sends a message 'message' through the socket.  If the send fails
        // send returns false.  Use getErrno or getError to retrieve the
        // error.
        bool send(const std::string &message);
        
        // Sends a message 'message' through the socket.  size is the number
        // of bytes from message that should be sent. If the send fails
        // send returns false.  Use getErrno or getError to retrieve the
        // error.
        bool send(const char *message, int size);
        
        // Recieves a message through the socket.  The message is put into
        // message.  If an error occurs recv returns false.  Use getErrno or
        // getError to retrieve the error.
        // If the connect was closed, recv returns false and sets the closed
        // attribute to true.
        bool recv(std::string *message);
        
        // Recieves a message through the socket and returns it.  If status
        // is nonzero it is set to false if an error occurs or true if there
        // is no error.  Use getErrno or getError to retrieve the error.
        // If the connect was closed, status is set to false and recv sets the
        // closed attribute to true.
        std::string recv(bool *status = 0);
        
        // Recieves a message through the socket.  The message is put into
        // message, placing at most 'avail' bytes into it.  If used is
        // non-zero it will be set to the number of bytes actually recieved.
        // If an error occurs recv returns -1, otherwise it returns the number
        // of bytes read into message. Use getErrno or getError to retrieve
        // the error.
        // If the connect was closed, recv returns 0 and sets the closed
        // attribute to true.
        int recv(char *message, int avail, int *used = 0);
        
        // Recieves a message through the socket.  The message is put into
        // message, placing at most 'size' bytes into it.  'size' will be set
        // to the number of bytes actually read into message.
        // If an error occurs recv returns 0, otherwise it returns message.
        // Use getErrno or getError to retrieve the error.
        // If the connect was closed, recv returns 0 and sets the closed
        // attribute to true.
        char *recv(char *message, int *size);
        
        // 'ip' will be set to the ip address of the 'guy on the other side'
        // of the socket.  If 'port' is non-zero it will be set to the port
        // the other side of the socket has connected with.  Returns false on
        // error.  Use getErrno or getError to retrieve the error.
        bool getpeername(std::string *ip, int *port = 0);
        
        // Returns the ip address of the 'guy on the other side' of the socket
        // in string form.  If port is non-zero it will be set to the port the
        // other side of the socket has connected with.  If status is non-zero
        // it will be set to false if there is was error or true if there was
        // not.  Use getErrno or getError to retrieve the error.
        std::string getpeername(int *port = 0, bool *status = 0);
        
        // Don't modify below here
        
        // Set to false when a connection on this socket closes.  Defaults to
        // true
        bool closed;
        
        Sock sock;
        
        // True if this socket owns 'sock' (ie it should call close on it).
        mutable bool own;
    };
    
    // Looks through a number of sockets for one with data waiting to be recieved.
    // sock is set to the Socket that has data waiting to be recieved.  If there
    // is no socket with data waiting to be recieved select waits 'msec'
    // milliseconds for data to arrive.  If data arrives ready is set to true,
    // if no data arrives before the timeout data is set to false.
    //
    // If msec is less then 0, select will poll indefinietely.
    //
    // Returns false if an error occured or true if no error occured.  Use
    // getErrno or getError to retrieve the error.
    //
    // begin points to the begining of the array of Sockets (or .begin() iterator),
    // end points to just past the end of the array of Sockets (or .end()
    // iterator).  sock will be set to the socket that has data waiting.  ready
    // will be set to true if sock is set, false if there is an error or a timeout
    // occurs.
    bool select(Socket *begin, Socket *end, bool *ready, Socket *sock = 0, int msec = 0);
}

#endif
