#ifndef WORLD_H
#define WORLD_H

#include <map>
#include <string>
#include "city.h"
#include "items.h"
#include "resource.h"
#include "resources.h"
#include "building.h"
#include "buildings.h"
#include "train.h"
#include "trains.h"
#include "terminal.h"
#include "track.h"
#include "user.h"

extern class World {
public:
    
    // Attempts to remove amount resources from the resource specified by id.
    // amount is set to the number of resource actually taken.
    // player refers to the player trying to mine this resource
    // 
    // This function will fail if:
    // There are no resources left at this resource field - 1
    // There is no resource minner at this locaiton owned by this player - 2
    // there is no user with this id - 3
    // there is no resource with this id - 4
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int takeResources(unsigned int id, unsigned int player, int *amount);
    
    // Attempts to add a the building 'building' to the world's list of buildings.
    // 
    // This function will fail if:
    // the building is too far from the specified track piece - 5
    // the building is owned by someone different then the track piece owner - 6
    // the building owner does not have enough money to build this building - 7
    // you cannot build a building here - 8
    // there is no user with this id - 3
    // there is no track piece with this id - 15
    // there is no building of this type in the database - 18
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int addBuilding(const Building &building, unsigned int *id = 0);
    
    // Attempts to remove the building specified by id from the world's list
    // of buildings.  This can only be done if player is the owner of that
    // building
    // 
    // This function will fail if:
    // The building owner is different then player - 9
    // there is no building with this id - 10
    // there is no user with this id - 3
    // there is no building of this type in the database - 18
    // the track connected to this building is invalid - 20
    // the track connected to this building was not pointing to this buidling - 21
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int removeBuilding(unsigned int id, unsigned int player);
    
    // Attempts to add a the user 'user' to the world's list of users.
    // 
    // This function will fail if:
    // The username already exists - 11
    // username has bad characters - 19
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int addUser(User user);
    
    // Attempts to add a terminal 'terminal' to the world's list of terminals.
    // 
    // This function will fail if:
    // The terminal owner does not have enough money to build this terminal - 7
    // The city is out of terminal locations for this terminal - 12
    // there is no user with this id - 3
    // Invalid city id - 17
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int addTerminal(const Terminal &terminal);
    
    // Attempts to remove the track pointed to by id.  player must match the
    // owner of the terminal
    // 
    // This function will fail if:
    // there is no terminal with this id - 13
    // there is no user with this id - 3
    // The player does not own this terminal - 9
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int removeTerminal(unsigned int id, unsigned int player);
    
    // Attempts to add a track 'track' to the world's list of tracks
    // 
    // This function will fail if:
    // The track owner does not have enough money to build this track - 7
    // A track cannot be placed here - 14
    // there is no user with this id - 3
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int addTrack(const Track &track);
    
    // Attempts to remove the track specified by id.  player must match the owner
    // of the track specified by id.
    // 
    // This function will fail if:
    // there is no track with this id - 15
    // there is no user with this id - 3
    // The player does not own this track - 9
    //
    // Returns 0 on success or non-zero on failure.  You can lookup the return
    // code in err::list to get a string representation of what went wrong.
    int removeTrack(unsigned int id, unsigned int player);
    
    /** Databases (world objects that don't require / provide saving) **/
    // The global items database
    Items itemDb;
    
    // The global train database
    Trains trainDb;
    
    // The global resources database
    Resources resourceDb;
    
    // The global building database
    Buildings buildingDb;
    
    /** Loading and saving routines **/
    // Loads the cities hash table from the given filename
    void loadCities(const std::string &filename);
    
    // Saves the cities hash table to a given filename
    void saveCities(const std::string &filename) const;
    
    // City unique IDs as Key, City properties as Value
    std::map<unsigned int, City> cities;
    
    // Loads the resources hash table from the given filename
    void loadResources(const std::string &filename);
    
    // Saves the resources hash table to a given filename
    void saveResources(const std::string &filename) const;
    
    // Resource unique IDs as Key, Resource properties as Value
    std::map<unsigned int, Resource> resources;
    
    // Loads the building hash table from a given filename
    void loadBuildings(const std::string &filename);
    
    // Saves the building hash table to a given filename
    void saveBuildings(const std::string &filename) const;
    
    // Building unique IDs as Key, Building properties as Value
    std::map<unsigned int, Building> buildings;
    
    // Loads the building hash table from a given filename
    void loadTerminals(const std::string &filename);
    
    // Saves the building hash table to a given filename
    void saveTerminals(const std::string &filename) const;
    
    // Building unique IDs as Key, Building properties as Value
    std::map<unsigned int, Terminal> terminals;
    
    // Loads the track hash table from a given filename
    void loadTracks(const std::string &filename);
    
    // Saves the track hash table to a given filename
    void saveTracks(const std::string &filename) const;
    
    // Track unique IDs as Key, Track properties as Value
    std::map<unsigned int, Track> tracks;
    
    // Loads the train hash table from a given filename
    void loadTrains(const std::string &filename);
    
    // Saves the train hash table to a given filename
    void saveTrains(const std::string &filename) const;
    
    // Train unique IDs as Key, Train properties as Value
    std::map<unsigned int, Train> trains;
    
    // Loads the User hash table from a given filename
    void loadUsers(const std::string &filename);
    
    // Saves the user hash table to a given filename
    void saveUsers(const std::string &filename) const;
    
    // User unique IDs as Key, User properties as Value
    std::map<unsigned int, User> users;
} world;

#endif
