#include "socket.h"
#include "log.h"
#include <iostream>
#include <cerrno>
#include <cstring>

using namespace std;
using namespace net;

#ifdef WIN32
#ifndef socklen_t
#define socklen_t int
#endif
#endif

#define BACKLOG 10     // how many pending connections queue will hold

string net::getError(int err)
{
    string ret;
    
    getError(&ret, err);
    
    return ret;
}

void net::getError(string *error, int err)
{
#ifdef WIN32
    CHAR *lpMsgBuf;
    CHAR *lpDisplayBuf;
    DWORD dw = err;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    lpDisplayBuf = (CHAR*)LocalAlloc(LMEM_ZEROINIT,
        strlen(lpMsgBuf) + 40);
    wsprintf(lpDisplayBuf,
        "error %d: %s",
        dw, lpMsgBuf);
    
    *error = lpDisplayBuf;
    
    LocalFree(lpMsgBuf);
    LocalFree(lpDisplayBuf);
#else
    *error = strerror(err);
#endif
}

int net::getErrno()
{
#ifdef WIN32
    return WSAGetLastError();
#else
    return errno;
#endif
}

string net::getHError(int herrno)
{
#ifdef WIN32
    return getError(herrno);
#else
    return hstrerror(herrno);
#endif
}

void net::getHError(string *error, int herrno)
{
#ifdef WIN32
    getError(error, herrno);
#else
    *error = hstrerror(herrno);
#endif
}

int net::getHErrno()
{
#ifdef WIN32
    return getErrno();
#else
    return h_errno;
#endif
}

bool net::gethostname(std::string *hostname)
{
    char buf[512];
    
    if(0 != ::gethostname(buf, sizeof(buf)))
        return false;
    
    *hostname = buf;
    
    return true;
}

string net::gethostname(bool *status)
{
    char buf[512];
    
    if(0 != ::gethostname(buf, sizeof(buf))) {
        
        if(status)
            *status = false;
        
        return string();
    }
    
    if(status)
        *status = true;
    
    return buf;
}

bool net::gethostbyname(const string &name, string *ip)
{
    struct hostent *h = ::gethostbyname("test");
    
    if(!h)
        return false;
    
    *ip = inet_ntoa(*((struct in_addr *)h->h_addr));
    
    return true;
}

string net::gethostbyname(const string &name, bool *status)
{
    struct hostent *h = ::gethostbyname(name.c_str());
    
    if(!h) {
        
        if(status)
            *status = false;
        
        return string();
    }
    
    if(status)
        *status = true;
    
    return inet_ntoa(*((struct in_addr *)h->h_addr));
}

bool net::gethostbyaddr(const std::string &ip, std::string *hostname)
{
    struct hostent *h = ::gethostbyaddr(ip.c_str(), ip.size(), AF_INET);
    
    if(!h)
        return false;
    
    *hostname = h->h_name;
    
    return true;
}

std::string gethostbyaddr(const std::string &ip, bool *status = 0)
{
    struct hostent *h = ::gethostbyaddr(ip.c_str(), ip.size(), AF_INET);
    
    if(!h) {
        
        if(status)
            *status = false;
        
        return string();
    }
    
    if(status)
        *status = true;
    
    return h->h_name;
}

namespace {
    
    int instances = 0;
    
    void instanceAdd()
    {
        if(!instances++) {
            
#ifdef WIN32
            WSADATA wsaData;
            
            if(WSAStartup(MAKEWORD(1, 1), &wsaData) != 0)
                throw "WSAStartup failed!";
#endif
        }
    }
    
    void instanceDel()
    {
        if(!--instances) {
            
#ifdef WIN32
            WSACleanup();
#endif
        }
    }
}

void net::init()
{
    instanceAdd();
}

void net::shutdown()
{
    instanceDel();
}

Socket::Socket()
    : closed(false), sock(INVALID_SOCKET), own(false)
{
    
}

Socket::Socket(Protocol protocol) throw(Error)
    : closed(false), own(true)
{
    instanceAdd();
    
    int type;
    
    if(protocol == UDP)
        type = SOCK_DGRAM;
    else
        type = SOCK_STREAM;
    
    sock = socket(PF_INET, type, 0);
    
    if(sock == INVALID_SOCKET)
        throw Error();
}

Socket::Socket(const Socket &socket, bool invalidate)
    : closed(false), sock(socket.sock), own(false)
{
    instanceAdd();
    
    if(invalidate && socket.own) {
        
        own = true;
        socket.own = false;
    }
}

Socket::Socket(Sock socket, bool invalidate)
    : closed(false), sock(socket), own(invalidate)
{
    instanceAdd();
}

Socket::~Socket()
{
    instanceDel();
    
    if(own && sock != INVALID_SOCKET) {
#ifdef WIN32
        closesocket(sock);
#else
        close(sock);
#endif
    }
}

bool Socket::bind(unsigned short port)
{
    struct sockaddr_in my_addr;

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);
    my_addr.sin_addr.s_addr = INADDR_ANY;
    memset(&(my_addr.sin_zero), '\0', 8);

    return 0 == ::bind(sock, (struct sockaddr *)&my_addr, sizeof(struct sockaddr));
}

bool Socket::listen()
{
    return 0 == ::listen(sock, BACKLOG);
}

bool Socket::connectIp(const string &ip, unsigned short port)
{
    struct sockaddr_in dest_addr;   // will hold the destination addr

    dest_addr.sin_family = AF_INET;
    dest_addr.sin_port = htons(port);
    dest_addr.sin_addr.s_addr = inet_addr(ip.c_str());
    memset(&(dest_addr.sin_zero), '\0', 8);
    
    return 0 == ::connect(sock, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr));
}

bool Socket::connect(const string &hostname, unsigned short port)
{
    string ip;
    
    if(!gethostbyname(hostname, &ip))
        return false;
    
    return connectIp(ip, port);
}

Socket::Sock Socket::acceptBsd(string *ip, unsigned short *port)
{
    struct sockaddr_in their_addr;
    socklen_t size = sizeof(struct sockaddr_in);
    Sock ret = ::accept(sock, (struct sockaddr *)&their_addr, &size);
    
    if(ret == INVALID_SOCKET)
        return INVALID_SOCKET;
    
    if(ip)
        *ip = inet_ntoa(their_addr.sin_addr);
    
    if(port)
        *port = ntohs(their_addr.sin_port);
    
    return ret;
}

Socket Socket::accept(bool *status, string *ip, unsigned short *port)
{
    Sock s = acceptBsd(ip, port);
    
    if(s == INVALID_SOCKET)
        *status = false;
    else
        *status = true;
    
    return Socket(s, *status);
}

bool Socket::send(const string &msg)
{
    return send(msg.c_str(), msg.size());
}

bool Socket::send(const char *message, int size)
{
    const char *itr = message;
    const char *end = message + size;
    int n;
    
    while(itr < end) {
        
        n = ::send(sock, itr, end - itr, 0);
        
        if(n == SOCKET_ERROR) {
            
            return false;
        }
        
        itr += n;
    }
    
    return true;
}

bool Socket::recv(std::string *message)
{
    char buf[1024];
    int size = sizeof(buf);
    
    if(0 == recv(buf, &size))
        return false;
    
    message->resize(size);
    
    for(int i = 0; i < size; i++) {
        
        (*message)[i] = buf[i];
    }
    
    return true;
}

string Socket::recv(bool *status)
{
    string ret;
    
    if(!recv(&ret)) {
        if(status)
            *status = false;
    }
    else if(status)
        *status = true;
    
    return ret;
}

int Socket::recv(char *message, int avail, int *used)
{
    int total = ::recv(sock, message, avail, 0);
    
    if(total == SOCKET_ERROR)
        total = -1;
    
    if(!total)
        closed = true;
    
    if(used)
        *used = total;
    
    return total;
}

char *Socket::recv(char *message, int *size)
{
    *size = ::recv(sock, message, *size, 0);
    
    if(*size == SOCKET_ERROR)
        return 0;
    
    if(!*size)
        closed = true;
    
    return message;
}

bool Socket::getpeername(string *ip, int *port)
{
    struct sockaddr_in addr;
    socklen_t len = sizeof(struct sockaddr_in);
    
    if(0 != ::getpeername(sock, (struct sockaddr*)&addr, &len))
        return false;
    
    *ip = inet_ntoa(addr.sin_addr);
    
    if(port)
        *port = ntohs(addr.sin_port);
    
    return true;
}

string Socket::getpeername(int *port, bool *status)
{
    struct sockaddr_in addr;
    socklen_t len = sizeof(struct sockaddr_in);
    
    if(0 != ::getpeername(sock, (struct sockaddr*)&addr, &len)) {
        
        if(status)
            *status = false;
        
        return string();
    }
    
    if(status)
        *status = false;
    
    if(port)
        *port = ntohs(addr.sin_port);
    
    return inet_ntoa(addr.sin_addr);
}

bool net::select(Socket *begin, Socket *end, bool *ready, Socket *sock, int msec)
{
    fd_set readfds;
    int numfds = 0;
    struct timeval tv = { 0, 0 };
    
    *ready = false;
    
    if(begin >= end)
        return false;
    
    tv.tv_sec = msec / 1000;
    tv.tv_usec = (msec % 1000) * 1000;
    
    FD_ZERO(&readfds);
    
    for(Socket *itr = begin; itr != end; ++itr) {
        
        if(signed(itr->sock + 1) > numfds)
            numfds = itr->sock + 1;
        
        FD_SET(itr->sock, &readfds);
    }
    
    switch(select(numfds, &readfds, 0, 0, &tv)) {
        case SOCKET_ERROR: return false;
        case 0: return true;
    }
    
    for(Socket *itr = begin; itr != end; ++itr) {
        
        if(FD_ISSET(itr->sock, &readfds)) {
            
            if(sock) {
                
                sock->sock = itr->sock;
                sock->own = false;
            }
            
            *ready = true;
            break;
        }
    }
    
    return true;
}
