#include "software.h"
#include "log.h"
#include <string>
using namespace std;

BITMAP *SoftwareRenderer::screen = 0;
bool SoftwareRenderer::ready = false;

static int instance_count = 0;
SoftwareRenderer softwareRenderer;

SoftwareRenderer::SoftwareRenderer() throw(const char*)
    : buffer(0), mode(def::mode), screen_w(def::w), screen_h(def::h)
{
	if(instance_count++)
		throw "Multiple conf instances are not allowed.";
}

int SoftwareRenderer::set_gfx_mode(int mode, int w, int h) {
    
	int ret = ::set_gfx_mode(mode, w, h, 0, 0);
	
	if(ret == 0) {
		screen = ::screen;
		this->mode = mode;
		screen_w = SCREEN_W;
		screen_h = SCREEN_H;
		
        if(screen)
            clear_to_color(screen, makecol(0, 0, 0));
	}
	
    ready = ret == 0;
    
	return ret;
}

bool SoftwareRenderer::init_page_flipping() {
	
	if(!(gfx_capabilities & GFX_CAN_TRIPLE_BUFFER))
		enable_triple_buffer();
    
    pagecount = 2;
    
    if(gfx_capabilities & GFX_CAN_TRIPLE_BUFFER)
        pagecount = 3;
    
    // This OS must specify requested video memory in set_gfx_mode
#ifdef ALLEGRO_VRAM_SINGLE_SURFACE
    if(0 != ::set_gfx_mode(mode, screen_w, screen_h, screen_w * pagecount, screen_h)) {
        
        // These settings should already have been tested to work
        ::set_gfx_mode(mode, screen_w, screen_h, 0, 0);
    }
#endif
	
    for(int i = 0; i < pagecount; i++) {
        
        pages[i] = create_video_bitmap(screen_w, screen_h);
        
        if(!pages[i])
            break;
        
        if(i == pagecount - 1) {
            
            page = 0;
            screen = pages[0];
            return true;
        }
    }
	
    for(int i = 0; i < pagecount; i++) {
        
        destroy_bitmap(pages[i]);
        pages[i] = 0;
    }
    
    return false;
}

bool SoftwareRenderer::init_double_buffer() {
    
	destroy_bitmap(buffer);
	
	if((buffer = create_bitmap(screen_w, screen_h))) {
        
		screen = buffer;
		
		clear_to_color(buffer, makecol(0, 0, 0));
	}
	
	return buffer;
}

int SoftwareRenderer::get_render_mode()
{
    return update_mode;
}

const char *SoftwareRenderer::get_mode() {
    
	if(update_mode == page_flip) {
        
        if(pagecount == 2)
            return "Using mode 'Page Flipping'";
        
        if(pagecount == 3)
            return "Using mode 'Triple Buffer'";
    }
    
	if(update_mode == double_buffer)
		return "Using mode 'Double Buffer'";
    
	else
		return "Using mode 'Direct Drawing'";
}

int SoftwareRenderer::init(int mode, int w, int h, int update_mode)
{
    shutdown();
    
    if(0 != set_gfx_mode(mode, w, h)) {
		
        if(0 != set_gfx_mode(this->mode, screen_w, screen_h)) {
            
            return -1;
        }
    }
	
    if(mode != GFX_TEXT) {
        
        if(update_mode & page_flip)
            if(init_page_flipping()) {
                
                this->update_mode = page_flip;
                return 0;
            }
        
        if(update_mode & double_buffer)
            if(init_double_buffer()) {
                
                this->update_mode = double_buffer;
                return 0;
            }
    }
    
    if(!(update_mode & raw))
        return -1;
    else
        update_mode = raw;
    
    return 0;
}

void SoftwareRenderer::shutdown() {
    
	destroy_bitmap(buffer);
    buffer = 0;
    
    for(int i = 0; i < pagecount; i++) {
        
        destroy_bitmap(pages[i]);
        pages[i] = 0;
    }
    
    pagecount = 0;
}

void SoftwareRenderer::post_draw_2_page()
{
    if(++page >= pagecount)
        page = 0;
    
    if(show_video_bitmap(pages[page]) != 0)
        throw "request_video_bitmap() failed.";
    
    int drawto = page + 1;
    
    if(drawto >= pagecount)
        drawto = 0;
    
    screen = pages[drawto];
}

void SoftwareRenderer::post_draw_3_page()
{
    while(poll_scroll());
    
    if(++page >= pagecount)
        page = 0;
    
    if(request_video_bitmap(pages[page]) != 0)
        throw "request_video_bitmap() failed.";
    
    int drawto = page + 1;
    
    if(drawto >= pagecount)
        drawto = 0;
    
    screen = pages[drawto];
}

void SoftwareRenderer::post_draw() {
    
    gui_set_screen(this->screen);
    
    if(0 == screen || update_mode == raw)
        return;
    
	if(update_mode == double_buffer) {
        
		blit(buffer, ::screen, 0, 0, 0, 0, screen_w, screen_h);
	}
    
    else if(update_mode == page_flip) {
        
        if(pagecount == 3)
            post_draw_3_page();
        else if(pagecount == 2)
            post_draw_2_page();
        else
            throw "Invalid pagecount value for page flipping.";
    }
}
